#!/usr/bin/env python
"""Subversion object for libvcs.

.. todo::

    The follow are from saltstack/salt (Apache license):

    - [`SvnSync.get_revision_file`](libvcs.svn.SvnSync.get_revision_file)

    The following are pypa/pip (MIT license):

    - [`SvnSync.convert_pip_url`](libvcs.svn.SvnSync.convert_pip_url)
    - [`SvnSync.get_url`](libvcs.svn.SvnSync.get_url)
    - [`SvnSync.get_revision`](libvcs.svn.SvnSync.get_revision)
    - [`get_rev_options`](libvcs.svn.get_rev_options)
"""  # NOQA: E5
import logging
import os
import pathlib
import re
from typing import Any, Optional
from urllib import parse as urlparse

from libvcs._internal.run import run
from libvcs._internal.types import StrOrBytesPath, StrPath

from .base import BaseSync, VCSLocation, convert_pip_url as base_convert_pip_url

logger = logging.getLogger(__name__)


def convert_pip_url(pip_url: str) -> VCSLocation:
    # hotfix the URL scheme after removing svn+ from svn+ssh:// re-add it
    url, rev = base_convert_pip_url(pip_url)
    if url.startswith("ssh://"):
        url = "svn+" + url
    return VCSLocation(url=url, rev=rev)


class SvnSync(BaseSync):
    bin_name = "svn"
    schemes = ("svn", "svn+ssh", "svn+http", "svn+https", "svn+svn")

    def __init__(self, *, url: str, dir: StrPath, **kwargs: Any) -> None:
        """A svn repository.

        Parameters
        ----------
        url : str
            URL in subversion repository

        svn_username : str, optional
            username to use for checkout and update

        svn_password : str, optional
            password to use for checkout and update

        svn_trust_cert : bool
            trust the Subversion server site certificate, default False
        """
        if "svn_trust_cert" not in kwargs:
            self.svn_trust_cert = False

        self.rev = kwargs.get("rev")
        super().__init__(url=url, dir=dir, **kwargs)

    def _user_pw_args(self) -> list[Any]:
        args = []
        for param_name in ["svn_username", "svn_password"]:
            if hasattr(self, param_name):
                args.extend(["--" + param_name[4:], getattr(self, param_name)])
        return args

    def obtain(self, quiet: Optional[bool] = None, *args: Any, **kwargs: Any) -> None:
        self.ensure_dir()

        url, rev = self.url, self.rev

        cmd: list[StrOrBytesPath] = ["checkout", "-q", url, "--non-interactive"]
        if self.svn_trust_cert:
            cmd.append("--trust-server-cert")
        cmd.extend(self._user_pw_args())
        cmd.extend(get_rev_options(url, rev))
        cmd.append(self.dir)

        self.run(cmd)

    def get_revision_file(self, location: str) -> int:
        """Return revision for a file."""

        current_rev = self.run(["info", location])

        _INI_RE = re.compile(r"^([^:]+):\s+(\S.*)$", re.M)

        info_list = _INI_RE.findall(current_rev)
        return int(dict(info_list)["Revision"])

    def get_revision(self, location: Optional[str] = None) -> int:
        """Return maximum revision for all files under a given location"""

        if not location:
            location = self.url

        if os.path.exists(location) and not os.path.isdir(location):
            return self.get_revision_file(location)

        # Note: taken from setuptools.command.egg_info
        revision = 0

        for base, dirs, files in os.walk(location):
            if ".svn" not in dirs:
                dirs[:] = []
                continue  # no sense walking uncontrolled subdirs
            dirs.remove(".svn")
            entries_fn = os.path.join(base, ".svn", "entries")
            if not os.path.exists(entries_fn):
                # FIXME: should we warn?
                continue

            dirurl, localrev = SvnSync._get_svn_url_rev(base)

            if base == location:
                assert dirurl is not None
                base = dirurl + "/"  # save the root url
            elif not dirurl or not dirurl.startswith(base):
                dirs[:] = []
                continue  # not part of the same svn tree, skip it
            revision = max(revision, localrev)
        return revision

    def update_repo(
        self, dest: Optional[str] = None, *args: Any, **kwargs: Any
    ) -> None:
        self.ensure_dir()
        if pathlib.Path(self.dir / ".svn").exists():
            if dest is None:
                dest = str(self.dir)

            url, rev = self.url, self.rev

            cmd = ["update"]
            cmd.extend(self._user_pw_args())
            cmd.extend(get_rev_options(url, rev))
            cmd.append(dest)

            self.run(cmd)
        else:
            self.obtain()
            self.update_repo()

    @classmethod
    def _get_svn_url_rev(cls, location: str) -> tuple[Optional[str], int]:
        _svn_xml_url_re = re.compile('url="([^"]+)"')
        _svn_rev_re = re.compile(r'committed-rev="(\d+)"')
        _svn_info_xml_rev_re = re.compile(r'\s*revision="(\d+)"')
        _svn_info_xml_url_re = re.compile(r"<url>(.*)</url>")

        entries_path = os.path.join(location, ".svn", "entries")
        if os.path.exists(entries_path):
            with open(entries_path) as f:
                data = f.read()
        else:  # subversion >= 1.7 does not have the 'entries' file
            data = ""

        url = None
        if data.startswith("8") or data.startswith("9") or data.startswith("10"):
            entries = list(map(str.splitlines, data.split("\n\x0c\n")))
            del entries[0][0]  # get rid of the '8'
            url = entries[0][3]
            revs = [int(d[9]) for d in entries if len(d) > 9 and d[9]] + [0]
        elif data.startswith("<?xml"):
            match = _svn_xml_url_re.search(data)
            if not match:
                raise ValueError(f"Badly formatted data: {data!r}")
            url = match.group(1)  # get repository URL
            revs = [int(m.group(1)) for m in _svn_rev_re.finditer(data)] + [0]
        else:
            try:
                # subversion >= 1.7
                # Note that using get_remote_call_options is not necessary here
                # because `svn info` is being run against a local directory.
                # We don't need to worry about making sure interactive mode
                # is being used to prompt for passwords, because passwords
                # are only potentially needed for remote server requests.
                xml = run(
                    ["svn", "info", "--xml", location],
                )
                match = _svn_info_xml_url_re.search(xml)
                assert match is not None
                url = match.group(1)
                revs = [int(m.group(1)) for m in _svn_info_xml_rev_re.finditer(xml)]
            except Exception:
                url, revs = None, []

        if revs:
            rev = max(revs)
        else:
            rev = 0

        return url, rev


def get_rev_options(url: str, rev: None) -> list[Any]:
    """Return revision options. From pip pip.vcs.subversion."""
    if rev:
        rev_options = ["-r", rev]
    else:
        rev_options = []

    r = urlparse.urlsplit(url)
    if hasattr(r, "username"):
        # >= Python-2.5
        username, password = r.username, r.password
    else:
        netloc = r[1]
        if "@" in netloc:
            auth = netloc.split("@")[0]
            if ":" in auth:
                username, password = auth.split(":", 1)
            else:
                username, password = auth, None
        else:
            username, password = None, None

    if username:
        rev_options += ["--username", username]
    if password:
        rev_options += ["--password", password]
    return rev_options



import pathlib
import typing as t

import pytest

from _pytest.doctest import DoctestItem


@pytest.fixture(autouse=True)
def add_doctest_fixtures(
    request: pytest.FixtureRequest,
    doctest_namespace: t.Dict[str, t.Any],
) -> None:
    if isinstance(request._pyfuncitem, DoctestItem):
        request.getfixturevalue("add_doctest_fixtures")
        request.getfixturevalue("set_home")


@pytest.fixture(autouse=True)
def setup(
    request: pytest.FixtureRequest,
    gitconfig: pathlib.Path,
    set_home: pathlib.Path,
) -> None:
    pass



"""tests for libvcs exceptions."""
import pytest

from libvcs import exc


def test_command_error() -> None:
    command = None
    with pytest.raises(exc.CommandError) as e:
        returncode = 0
        command = ["command", "arg"]
        raise exc.CommandError("this is output", returncode, command)

    assert command is not None

    assert e.value.cmd == " ".join(command)
    assert (
        str(e.value)
        == exc.CommandError.message.format(
            returncode=e.value.returncode, cmd=e.value.cmd
        )
        + "\n%s" % e.value.output
    )

    with pytest.raises(exc.CommandError) as e:
        returncode = 0
        command = ["command", "arg"]
        raise exc.CommandError("", returncode, command)
    assert e.value.cmd == " ".join(command)
    assert str(e.value) == exc.CommandError.message.format(
        returncode=e.value.returncode, cmd=e.value.cmd
    )

    command_2 = None

    with pytest.raises(exc.CommandError) as e:
        command_2 = "command arg"
        raise exc.CommandError("this is output", 0, command_2)

    assert command_2 is not None
    assert e.value.cmd == command_2



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['libvcs', 'libvcs._internal', 'libvcs.cmd', 'libvcs.sync', 'libvcs.url']

package_data = \
{'': ['*']}

extras_require = \
{':python_version == "3.10"': ['typing-extensions>=4.1.1,<5.0.0']}

entry_points = \
{'pytest11': ['libvcs = libvcs.pytest_plugin']}

setup_kwargs = {
    'name': 'libvcs',
    'version': '0.16.3',
    'description': 'Lite, typed, python utilities for Git, SVN, Mercurial, etc.',
    'long_description': '# `libvcs` &middot; [![Python Package](https://img.shields.io/pypi/v/libvcs.svg)](https://pypi.org/project/libvcs/) [![License](https://img.shields.io/github/license/vcs-python/libvcs.svg)](https://github.com/vcs-python/libvcs/blob/master/LICENSE) [![Code Coverage](https://codecov.io/gh/vcs-python/libvcs/branch/master/graph/badge.svg)](https://codecov.io/gh/vcs-python/libvcs)\n\nlibvcs is a lite, [typed](https://docs.python.org/3/library/typing.html), pythonic tool box for\ndetection and parsing of URLs, commanding, and syncing with `git`, `hg`, and `svn`. Powers\n[vcspull](https://www.github.com/vcs-python/vcspull/).\n\n## Overview\n\nFeatures for Git, Subversion, and Mercurial:\n\n- **Detect and parse** VCS URLs\n- **Command** VCS via python API\n- **Sync** repos locally\n- **Test fixtures** for temporary local repos and working copies\n\nTo **get started**, see the [quickstart](https://libvcs.git-pull.com/quickstart.html) for more.\n\n```console\n$ pip install --user libvcs\n```\n\n## URL Parser\n\nYou can validate and parse Git, Mercurial, and Subversion URLs through\n[`libvcs.url`](https://libvcs.git-pull.com/url/index.html):\n\nValidate:\n\n```python\n>>> from libvcs.url.git import GitURL\n\n>>> GitURL.is_valid(url=\'https://github.com/vcs-python/libvcs.git\')\nTrue\n```\n\nParse and adjust a Git URL:\n\n```python\n>>> from libvcs.url.git import GitURL\n\n>>> git_location = GitURL(url=\'git@github.com:vcs-python/libvcs.git\')\n\n>>> git_location\nGitURL(url=git@github.com:vcs-python/libvcs.git,\n        user=git,\n        hostname=github.com,\n        path=vcs-python/libvcs,\n        suffix=.git,\n        matcher=core-git-scp)\n```\n\nSwitch repo libvcs -> vcspull:\n\n```python\n>>> from libvcs.url.git import GitURL\n\n>>> git_location = GitURL(url=\'git@github.com:vcs-python/libvcs.git\')\n\n>>> git_location.path = \'vcs-python/vcspull\'\n\n>>> git_location.to_url()\n\'git@github.com:vcs-python/vcspull.git\'\n\n# Switch them to gitlab:\n>>> git_location.hostname = \'gitlab.com\'\n\n# Export to a `git clone` compatible URL.\n>>> git_location.to_url()\n\'git@gitlab.com:vcs-python/vcspull.git\'\n```\n\nSee more in the [parser document](https://libvcs.git-pull.com/parse/index.html).\n\n## Commands\n\nSimple [`subprocess`](https://docs.python.org/3/library/subprocess.html) wrappers around `git(1)`,\n`hg(1)`, `svn(1)`. Here is [`Git`](https://libvcs.git-pull.com/cmd/git.html#libvcs.cmd.git.Git) w/\n[`Git.clone`](http://libvcs.git-pull.com/cmd/git.html#libvcs.cmd.git.Git.clone):\n\n```python\nimport pathlib\nfrom libvcs.cmd.git import Git\n\ngit = Git(dir=pathlib.Path.cwd() / \'my_git_repo\')\ngit.clone(url=\'https://github.com/vcs-python/libvcs.git\')\n```\n\n## Sync\n\nCreate a [`GitSync`](https://libvcs.git-pull.com/projects/git.html#libvcs.sync.git.GitProject)\nobject of the project to inspect / checkout / update:\n\n```python\nimport pathlib\nfrom libvcs.sync.git import GitSync\n\nrepo = GitSync(\n   url="https://github.com/vcs-python/libvcs",\n   dir=pathlib.Path().cwd() / "my_repo",\n   remotes={\n       \'gitlab\': \'https://gitlab.com/vcs-python/libvcs\'\n   }\n)\n\n# Update / clone repo:\n>>> repo.update_repo()\n\n# Get revision:\n>>> repo.get_revision()\nu\'5c227e6ab4aab44bf097da2e088b0ff947370ab8\'\n```\n\n## Pytest plugin\n\nlibvcs also provides a test rig for local repositories. It automatically can provide clean local\nrepositories and working copies for git, svn, and mercurial. They are automatically cleaned up after\neach test.\n\nIt works by bootstrapping a temporary `$HOME` environment in a\n[`TmpPathFactory`](https://docs.pytest.org/en/7.1.x/reference/reference.html#tmp-path-factory-factory-api)\nfor automatic cleanup.\n\n```python\nimport pathlib\n\nfrom libvcs.pytest_plugin import CreateProjectCallbackFixtureProtocol\nfrom libvcs.sync.git import GitSync\n\n\ndef test_repo_git_remote_checkout(\n    create_git_remote_repo: CreateProjectCallbackFixtureProtocol,\n    tmp_path: pathlib.Path,\n    projects_path: pathlib.Path,\n) -> None:\n    git_server = create_git_remote_repo()\n    git_repo_checkout_dir = projects_path / "my_git_checkout"\n    git_repo = GitSync(dir=str(git_repo_checkout_dir), url=f"file://{git_server!s}")\n\n    git_repo.obtain()\n    git_repo.update_repo()\n\n    assert git_repo.get_revision() == "initial"\n\n    assert git_repo_checkout_dir.exists()\n    assert pathlib.Path(git_repo_checkout_dir / ".git").exists()\n```\n\nLearn more on the docs at https://libvcs.git-pull.com/pytest-plugin.html\n\n## Donations\n\nYour donations fund development of new features, testing and support. Your money will go directly to\nmaintenance and development of the project. If you are an individual, feel free to give whatever\nfeels right for the value you get out of the project.\n\nSee donation options at <https://www.git-pull.com/support.html>.\n\n## More information\n\n- Python support: 3.9+, pypy\n- VCS supported: git(1), svn(1), hg(1)\n- Source: <https://github.com/vcs-python/libvcs>\n- Docs: <https://libvcs.git-pull.com>\n- Changelog: <https://libvcs.git-pull.com/history.html>\n- APIs for git, hg, and svn:\n  - [`libvcs.url`](https://libvcs.git-pull.com/url/): URL Parser\n  - [`libvcs.cmd`](https://libvcs.git-pull.com/cmd/): Commands\n  - [`libvcs.sync`](https://libvcs.git-pull.com/sync/): Clone and update\n- Issues: <https://github.com/vcs-python/libvcs/issues>\n- Test Coverage: <https://codecov.io/gh/vcs-python/libvcs>\n- pypi: <https://pypi.python.org/pypi/libvcs>\n- Open Hub: <https://www.openhub.net/p/libvcs>\n- License: [MIT](https://opensource.org/licenses/MIT).\n\n[![Docs](https://github.com/vcs-python/libvcs/workflows/docs/badge.svg)](https://libvcs.git-pull.com/)\n[![Build Status](https://github.com/vcs-python/libvcs/workflows/tests/badge.svg)](https://github.com/vcs-python/libvcs/actions?query=workflow%3A%22tests%22)\n',
    'author': 'Tony Narlock',
    'author_email': 'tony@git-pull.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'http://github.com/vcs-python/libvcs/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""This module is an all-in-one parser and validator for Mercurial URLs.

- Detection: :meth:`HgURL.is_valid()`
- Parse: :class:`HgURL`

  compare to :class:`urllib.parse.ParseResult`

  - Output ``hg(1)`` URL: :meth:`HgURL.to_url()`
- Extendable via :class:`~libvcs.url.base.MatcherRegistry`,
  :class:`~libvcs.url.base.Matcher`

.. Note::

   Do you use Mercurial at your job or project? This module welcomes a champion /
   maintainer to assure support is top-tier. Stop by the `project tracker
   <https://github.com/vcs-python/libvcs>`_ to make yourself known. We won't stabilize
   any APIs until we're satisfied support is "by the book" and is bullet proofed.
"""

import dataclasses
import re
from typing import Optional

from libvcs._internal.dataclasses import SkipDefaultFieldsReprMixin

from .base import Matcher, MatcherRegistry, URLProtocol

RE_PATH = r"""
    ((?P<user>\w+)@)?
    (?P<hostname>([^/:@]+))
    (:(?P<port>\d{1,5}))?
    (?P<separator>/)?
    (?P<path>
      /?(\w[^:.]*)
    )?
"""

RE_SCHEME = r"""
    (?P<scheme>
      (
        http|https|ssh
      )
    )
"""

DEFAULT_MATCHERS: list[Matcher] = [
    Matcher(
        label="core-hg",
        description="Vanilla hg pattern",
        pattern=re.compile(
            rf"""
        ^{RE_SCHEME}
        ://
        {RE_PATH}
        """,
            re.VERBOSE,
        ),
    ),
]
"""Core regular expressions. These are patterns understood by ``hg(1)``"""


#
# Third-party URLs, e.g. pip, etc.
#
RE_PIP_SCHEME = r"""
    (?P<scheme>
      (
        hg\+ssh|
        hg\+https|
        hg\+http|
        hg\+file
      )
    )
"""

PIP_DEFAULT_MATCHERS: list[Matcher] = [
    Matcher(
        label="pip-url",
        description="pip-style hg URL",
        pattern=re.compile(
            rf"""
        {RE_PIP_SCHEME}
        ://
        {RE_PATH}
        """,
            re.VERBOSE,
        ),
    ),
    # file://, RTC 8089, File:// https://datatracker.ietf.org/doc/html/rfc8089
    Matcher(
        label="pip-file-url",
        description="pip-style hg+file:// URL",
        pattern=re.compile(
            r"""
        (?P<scheme>hg\+file)://
        (?P<path>.*)
        """,
            re.VERBOSE,
        ),
    ),
]
"""pip-style hg URLs.

Examples of PIP-style hg URLs (via pip.pypa.io)::

    MyProject @ hg+http://hg.myproject.org/MyProject
    MyProject @ hg+https://hg.myproject.org/MyProject
    MyProject @ hg+ssh://hg.myproject.org/MyProject
    MyProject @ hg+file:///home/user/projects/MyProject

Refs (via pip.pypa.io)::

    MyProject @ hg+http://hg.example.com/MyProject@da39a3ee5e6b
    MyProject @ hg+http://hg.example.com/MyProject@2019
    MyProject @ hg+http://hg.example.com/MyProject@v1.0
    MyProject @ hg+http://hg.example.com/MyProject@special_feature

Notes
-----

- https://pip.pypa.io/en/stable/topics/vcs-support/
"""  # NOQA: E501


@dataclasses.dataclass(repr=False)
class HgURL(URLProtocol, SkipDefaultFieldsReprMixin):
    """Mercurial repository location. Parses URLs on initialization.

    Attributes
    ----------
    matcher : str
        name of the :class:`~libvcs.url.base.Matcher`

    Examples
    --------
    >>> HgURL(url='https://hg.mozilla.org/mozilla-central/')
    HgURL(url=https://hg.mozilla.org/mozilla-central/,
            scheme=https,
            hostname=hg.mozilla.org,
            path=mozilla-central/,
            matcher=core-hg)

    >>> myrepo = HgURL(url='https://hg.mozilla.org/mozilla-central/')

    >>> myrepo.hostname
    'hg.mozilla.org'

    >>> myrepo.path
    'mozilla-central/'

    >>> HgURL(url='ssh://username@machinename/path/to/repo')
    HgURL(url=ssh://username@machinename/path/to/repo,
            scheme=ssh,
            user=username,
            hostname=machinename,
            path=path/to/repo,
            matcher=core-hg)

    - Compatibility checking: :meth:`HgURL.is_valid()`
    - URLs compatible with ``hg(1)``: :meth:`HgURL.to_url()`
    """

    url: str
    scheme: Optional[str] = None
    user: Optional[str] = None
    hostname: str = dataclasses.field(default="")
    port: Optional[int] = None
    separator: str = dataclasses.field(default="/")
    path: str = dataclasses.field(default="")

    #
    # commit-ish: tag, branch, ref, revision
    #
    ref: Optional[str] = None

    matcher: Optional[str] = None
    # name of the :class:`Matcher`
    matchers: MatcherRegistry = MatcherRegistry(
        _matchers={m.label: m for m in DEFAULT_MATCHERS}
    )

    def __post_init__(self) -> None:
        url = self.url
        for matcher in self.matchers.values():
            match = re.match(matcher.pattern, url)
            if match is None:
                continue
            groups = match.groupdict()
            setattr(self, "matcher", matcher.label)
            for k, v in groups.items():
                setattr(self, k, v)

            for k, v in matcher.pattern_defaults.items():
                if getattr(self, k, None) is None:
                    setattr(self, k, matcher.pattern_defaults[k])

    @classmethod
    def is_valid(cls, url: str, is_explicit: Optional[bool] = False) -> bool:
        """Whether URL is compatible with VCS or not.

        Examples
        --------

        >>> HgURL.is_valid(
        ...     url='https://hg.mozilla.org/mozilla-central'
        ... )
        True

        >>> HgURL.is_valid(url='ssh://hg@hg.python.org/cpython')
        True

        >>> HgURL.is_valid(url='notaurl')
        False
        """
        return any(re.search(matcher.pattern, url) for matcher in cls.matchers.values())

    def to_url(self) -> str:
        """Return a ``hg(1)``-compatible URL. Can be used with ``hg clone``.

        Examples
        --------

        >>> hg_url = HgURL(url='https://hg.mozilla.org/mozilla-central')

        >>> hg_url
        HgURL(url=https://hg.mozilla.org/mozilla-central,
                scheme=https,
                hostname=hg.mozilla.org,
                path=mozilla-central,
                matcher=core-hg)

        Switch repo libvcs -> vcspull:

        >>> hg_url.path = 'mobile-browser'

        >>> hg_url.to_url()
        'https://hg.mozilla.org/mobile-browser'

        Switch them to localhost:

        >>> hg_url.hostname = 'localhost'
        >>> hg_url.scheme = 'http'

        >>> hg_url.to_url()
        'http://localhost/mobile-browser'

        Another example, `hugin <http://hugin.hg.sourceforge.net>`_:

        >>> hugin = HgURL(url="http://hugin.hg.sourceforge.net:8000/hgroot/hugin/hugin")

        >>> hugin
        HgURL(url=http://hugin.hg.sourceforge.net:8000/hgroot/hugin/hugin,
                scheme=http,
                hostname=hugin.hg.sourceforge.net,
                port=8000,
                path=hgroot/hugin/hugin,
                matcher=core-hg)

        >>> hugin.to_url()
        'http://hugin.hg.sourceforge.net:8000/hgroot/hugin/hugin'

        SSH URL with a username, `graphicsmagic <http://graphicsmagick.org/Hg.html>`_:

        >>> graphicsmagick = HgURL(
        ...     url="ssh://yourid@hg.GraphicsMagick.org//hg/GraphicsMagick"
        ... )

        >>> graphicsmagick
        HgURL(url=ssh://yourid@hg.GraphicsMagick.org//hg/GraphicsMagick,
                scheme=ssh,
                user=yourid,
                hostname=hg.GraphicsMagick.org,
                path=/hg/GraphicsMagick,
                matcher=core-hg)

        >>> graphicsmagick.to_url()
        'ssh://yourid@hg.GraphicsMagick.org//hg/GraphicsMagick'

        Switch the username:

        >>> graphicsmagick.user = 'lucas'

        >>> graphicsmagick.to_url()
        'ssh://lucas@hg.GraphicsMagick.org//hg/GraphicsMagick'

        """
        parts = [self.scheme or "ssh", "://"]
        if self.user:
            parts.extend([self.user, "@"])

        parts.append(self.hostname)

        if self.port is not None:
            parts.extend([":", f"{self.port}"])

        parts.extend([self.separator, self.path])

        return "".join(part for part in parts if isinstance(part, str))



import typing

import pytest

from libvcs.sync.git import GitSync
from libvcs.url.base import MatcherRegistry
from libvcs.url.git import DEFAULT_MATCHERS, PIP_DEFAULT_MATCHERS, GitBaseURL, GitURL


class GitURLFixture(typing.NamedTuple):
    url: str
    is_valid: bool
    git_url: GitURL


TEST_FIXTURES: list[GitURLFixture] = [
    GitURLFixture(
        url="https://github.com/vcs-python/libvcs.git",
        is_valid=True,
        git_url=GitURL(
            url="https://github.com/vcs-python/libvcs.git",
            scheme="https",
            hostname="github.com",
            path="vcs-python/libvcs",
        ),
    ),
    GitURLFixture(
        url="https://github.com/vcs-python/libvcs",
        is_valid=True,
        git_url=GitURL(
            url="https://github.com/vcs-python/libvcs",
            scheme="https",
            hostname="github.com",
            path="vcs-python/libvcs",
        ),
    ),
    GitURLFixture(
        url="https://github.com:7999/vcs-python/libvcs",
        is_valid=True,
        git_url=GitURL(
            url="https://github.com:7999/vcs-python/libvcs",
            scheme="https",
            hostname="github.com",
            port=7999,
            path="vcs-python/libvcs",
        ),
    ),
    #
    # SCP-style URLs:
    # e.g. 'git@example.com:foo/bar.git'
    #
    GitURLFixture(
        url="git@github.com:liuxinyu95/AlgoXY.git",
        is_valid=True,
        git_url=GitURL(
            url="git@github.com:liuxinyu95/AlgoXY.git",
            scheme=None,
            hostname="github.com",
            path="liuxinyu95/AlgoXY",
        ),
    ),
    GitURLFixture(
        url="git@github.com:vcs-python/libvcs.git",
        is_valid=True,
        git_url=GitURL(
            url="git@github.com:vcs-python/libvcs.git",
            scheme="https",
            hostname="github.com",
            path="vcs-python/libvcs",
        ),
    ),
]


@pytest.mark.parametrize(
    "url,is_valid,git_url",
    TEST_FIXTURES,
)
def test_git_url(
    url: str,
    is_valid: bool,
    git_url: GitURL,
    git_repo: GitSync,
) -> None:
    url = url.format(local_repo=git_repo.dir)
    git_url.url = git_url.url.format(local_repo=git_repo.dir)

    assert GitURL.is_valid(url) == is_valid, f"{url} compatibility should be {is_valid}"
    assert GitURL(url) == git_url


class GitURLKwargs(typing.TypedDict):
    url: str


class GitURLKwargsFixture(typing.NamedTuple):
    url: str
    is_valid: bool
    git_url_kwargs: GitURLKwargs


#
#
# Extensibility: pip(1)
# w/ VCS prefixes, e.g. git+https, git+ssh, git+file
# https://pip.pypa.io/en/stable/topics/vcs-support/
#
#
PIP_TEST_FIXTURES: list[GitURLKwargsFixture] = [
    GitURLKwargsFixture(
        url="git+https://github.com/liuxinyu95/AlgoXY.git",
        is_valid=True,
        git_url_kwargs=GitURLKwargs(url="git+https://github.com/liuxinyu95/AlgoXY.git"),
    ),
    GitURLKwargsFixture(
        url="git+ssh://git@github.com:tony/AlgoXY.git",
        is_valid=True,
        git_url_kwargs=GitURLKwargs(url="git+ssh://git@github.com:tony/AlgoXY.git"),
    ),
    GitURLKwargsFixture(
        url="git+file://{local_repo}",
        is_valid=True,
        git_url_kwargs=GitURLKwargs(url="git+file://{local_repo}"),
    ),
    # Incompatible
    GitURLKwargsFixture(
        url="git+ssh://git@github.com/tony/AlgoXY.git",
        is_valid=True,
        git_url_kwargs=GitURLKwargs(url="git+ssh://git@github.com/tony/AlgoXY.git"),
    ),
]


@pytest.mark.parametrize(
    "url,is_valid,git_url_kwargs",
    PIP_TEST_FIXTURES,
)
def test_git_url_extension_pip(
    url: str,
    is_valid: bool,
    git_url_kwargs: GitURLKwargs,
    git_repo: GitSync,
) -> None:
    class GitURLWithPip(GitBaseURL):
        matchers: MatcherRegistry = MatcherRegistry(
            _matchers={m.label: m for m in [*DEFAULT_MATCHERS, *PIP_DEFAULT_MATCHERS]}
        )

    git_url_kwargs["url"] = git_url_kwargs["url"].format(local_repo=git_repo.dir)
    url = url.format(local_repo=git_repo.dir)
    git_url = GitURLWithPip(**git_url_kwargs)
    git_url.url = git_url.url.format(local_repo=git_repo.dir)

    assert (
        GitBaseURL.is_valid(url) != is_valid
    ), f"{url} compatibility should work with core, expects {not is_valid}"
    assert (
        GitURLWithPip.is_valid(url) == is_valid
    ), f"{url} compatibility should be {is_valid}"
    assert GitURLWithPip(url) == git_url


class ToURLFixture(typing.NamedTuple):
    git_url: GitURL
    expected: str


@pytest.mark.parametrize(
    "git_url,expected",
    [
        ToURLFixture(
            expected="https://github.com/vcs-python/libvcs.git",
            git_url=GitURL(
                url="https://github.com/vcs-python/libvcs.git",
                scheme="https",
                hostname="github.com",
                path="vcs-python/libvcs",
            ),
        ),
        ToURLFixture(
            expected="https://github.com/vcs-python/libvcs",
            git_url=GitURL(
                url="https://github.com/vcs-python/libvcs",
                scheme="https",
                hostname="github.com",
                path="vcs-python/libvcs",
            ),
        ),
        #
        # SCP-style URLs:
        # e.g. 'git@example.com:foo/bar.git'
        #
        ToURLFixture(
            expected="git@github.com:liuxinyu95/AlgoXY.git",
            git_url=GitURL(
                url="git@github.com:liuxinyu95/AlgoXY.git",
                scheme=None,
                hostname="github.com",
                path="liuxinyu95/AlgoXY",
            ),
        ),
        ToURLFixture(
            expected="git@github.com:vcs-python/libvcs.git",
            git_url=GitURL(
                url="git@github.com:vcs-python/libvcs.git",
                scheme="https",
                hostname="github.com",
                path="vcs-python/libvcs",
            ),
        ),
    ],
)
def test_git_to_url(
    expected: str,
    git_url: GitURL,
    git_repo: GitSync,
) -> None:
    """Test GitURL.to_url()"""
    git_url.url = git_url.url.format(local_repo=git_repo.dir)

    assert git_url.to_url() == expected


class RevFixture(typing.NamedTuple):
    git_url_kwargs: GitURLKwargs
    expected: typing.Optional[str]
    # Expected revision / branch / tag


@pytest.mark.parametrize(
    "git_url_kwargs,expected",
    [
        RevFixture(
            expected=None,
            git_url_kwargs=GitURLKwargs(
                url="git+ssh://git@bitbucket.example.com:7999/PROJ/repo.git",
            ),
        ),
        RevFixture(
            expected="eucalyptus",
            git_url_kwargs=GitURLKwargs(
                url="git+ssh://git@bitbucket.example.com:7999/PROJ/repo.git@eucalyptus",
            ),
        ),
        RevFixture(
            expected="build.2600-whistler",
            git_url_kwargs=GitURLKwargs(
                url="git+https://github.com/PROJ/repo.git@build.2600-whistler",
            ),
        ),
    ],
)
def test_git_revs(
    expected: str,
    git_url_kwargs: GitURLKwargs,
) -> None:
    class GitURLWithPip(GitURL):
        matchers: MatcherRegistry = MatcherRegistry(
            _matchers={m.label: m for m in [*DEFAULT_MATCHERS, *PIP_DEFAULT_MATCHERS]}
        )

    git_url = GitURLWithPip(**git_url_kwargs)
    assert git_url.rev == expected



import pathlib
from typing import Any, Callable, Union

import pytest

from libvcs.cmd import git


@pytest.mark.parametrize("dir_type", [str, pathlib.Path])
def test_run(
    dir_type: Callable[[Union[str, pathlib.Path]], Any], tmp_path: pathlib.Path
) -> None:
    repo = git.Git(dir=dir_type(tmp_path))

    assert repo.dir == tmp_path



import pathlib
from typing import Literal, Optional, TypedDict, TypeVar, Union

import pytest

from libvcs import GitSync, HgSync, SvnSync
from libvcs._internal.run import ProgressCallbackProtocol
from libvcs._internal.shortcuts import create_project
from libvcs._internal.types import StrPath
from libvcs.exc import InvalidVCS


class CreateProjectKwargsDict(TypedDict, total=False):
    url: str
    dir: StrPath
    vcs: Literal["git"]
    progress_callback: Optional[ProgressCallbackProtocol]


E = TypeVar("E", bound=BaseException)


@pytest.mark.parametrize(
    "repo_dict,repo_class,raises_exception",
    [
        (
            {"url": "https://github.com/freebsd/freebsd.git", "vcs": "git"},
            GitSync,
            None,
        ),
        (
            {"url": "https://bitbucket.org/birkenfeld/sphinx", "vcs": "hg"},
            HgSync,
            None,
        ),
        (
            {"url": "http://svn.code.sf.net/p/docutils/code/trunk", "vcs": "svn"},
            SvnSync,
            None,
        ),
        (
            {"url": "http://svn.code.sf.net/p/docutils/code/trunk", "vcs": "svna"},
            None,
            InvalidVCS,
        ),
    ],
)
def test_create_project(
    tmp_path: pathlib.Path,
    repo_dict: CreateProjectKwargsDict,
    repo_class: type[Union[SvnSync, GitSync, HgSync]],
    raises_exception: Union[None, Union[type[E], tuple[type[E], ...]]],
) -> None:
    # add parent_dir via fixture
    repo_dict["dir"] = tmp_path / "repo_name"

    if raises_exception is not None:
        with pytest.raises(raises_exception):
            create_project(**repo_dict)
    else:
        repo = create_project(**repo_dict)
        assert isinstance(repo, repo_class)



"""Utilities for filtering or searching :class:`list` of objects / list data.

Note
----
This is an internal API not covered by versioning policy.
"""
import re
import traceback
from collections.abc import Mapping, Sequence
from re import Pattern
from typing import Any, Callable, Optional, Protocol, TypeVar, Union

T = TypeVar("T", Any, Any)


def keygetter(
    obj: Mapping[str, Any],
    path: str,
) -> Union[None, Any, str, list[str], Mapping[str, str]]:
    """Fetch values in objects and keys, deeply.

    **With dictionaries**:

    >>> keygetter({ "menu": { "breakfast": "cereal" } }, "menu")
    {'breakfast': 'cereal'}

    >>> keygetter({ "menu": { "breakfast": "cereal" } }, "menu__breakfast")
    'cereal'

    **With objects**:

    >>> from typing import Optional
    >>> from dataclasses import dataclass, field

    >>> @dataclass()
    ... class Menu:
    ...     fruit: list[str] = field(default_factory=list)
    ...     breakfast: Optional[str] = None


    >>> @dataclass()
    ... class Restaurant:
    ...     place: str
    ...     city: str
    ...     state: str
    ...     menu: Menu = field(default_factory=Menu)


    >>> restaurant = Restaurant(
    ...     place="Largo",
    ...     city="Tampa",
    ...     state="Florida",
    ...     menu=Menu(
    ...         fruit=["banana", "orange"], breakfast="cereal"
    ...     )
    ... )

    >>> restaurant
    Restaurant(place='Largo',
        city='Tampa',
        state='Florida',
        menu=Menu(fruit=['banana', 'orange'], breakfast='cereal'))

    >>> keygetter(restaurant, "menu")
    Menu(fruit=['banana', 'orange'], breakfast='cereal')

    >>> keygetter(restaurant, "menu__breakfast")
    'cereal'
    """
    try:
        sub_fields = path.split("__")
        dct = obj
        for sub_field in sub_fields:
            if isinstance(dct, dict):
                dct = dct[sub_field]
            elif hasattr(dct, sub_field):
                dct = getattr(dct, sub_field)
        return dct
    except Exception as e:
        traceback.print_stack()
        print(f"Above error was {e}")
    return None


def parse_lookup(obj: Mapping[str, Any], path: str, lookup: str) -> Optional[Any]:
    """Check if field lookup key, e.g. "my__path__contains" has comparator, return val.

    If comparator not used or value not found, return None.

    >>> parse_lookup({ "food": "red apple" }, "food__istartswith", "__istartswith")
    'red apple'

    It can also look up objects:

    >>> from dataclasses import dataclass

    >>> @dataclass()
    ... class Inventory:
    ...     food: str

    >>> item = Inventory(food="red apple")

    >>> item
    Inventory(food='red apple')

    >>> parse_lookup(item, "food__istartswith", "__istartswith")
    'red apple'
    """
    try:
        if isinstance(path, str) and isinstance(lookup, str) and path.endswith(lookup):
            if field_name := path.rsplit(lookup)[0]:
                return keygetter(obj, field_name)
    except Exception:
        traceback.print_stack()
    return None


class LookupProtocol(Protocol):
    """Protocol for :class:`QueryList` filtering operators."""

    def __call__(
        self,
        data: Union[str, list[str], Mapping[str, str]],
        rhs: Union[str, list[str], Mapping[str, str], Pattern[str]],
    ) -> bool:
        """Callback for :class:`QueryList` filtering operators."""
        ...


def lookup_exact(
    data: Union[str, list[str], Mapping[str, str]],
    rhs: Union[str, list[str], Mapping[str, str], Pattern[str]],
) -> bool:
    return rhs == data


def lookup_iexact(
    data: Union[str, list[str], Mapping[str, str]],
    rhs: Union[str, list[str], Mapping[str, str], Pattern[str]],
) -> bool:
    if not isinstance(rhs, str) or not isinstance(data, str):
        return False

    return rhs.lower() == data.lower()


def lookup_contains(
    data: Union[str, list[str], Mapping[str, str]],
    rhs: Union[str, list[str], Mapping[str, str], Pattern[str]],
) -> bool:
    if not isinstance(rhs, str) or not isinstance(data, (str, Mapping, list)):
        return False

    return rhs in data


def lookup_icontains(
    data: Union[str, list[str], Mapping[str, str]],
    rhs: Union[str, list[str], Mapping[str, str], Pattern[str]],
) -> bool:
    if not isinstance(rhs, str) or not isinstance(data, (str, Mapping, list)):
        return False

    if isinstance(data, str):
        return rhs.lower() in data.lower()
    if isinstance(data, Mapping):
        return rhs.lower() in [k.lower() for k in data.keys()]

    return False


def lookup_startswith(
    data: Union[str, list[str], Mapping[str, str]],
    rhs: Union[str, list[str], Mapping[str, str], Pattern[str]],
) -> bool:
    if not isinstance(rhs, str) or not isinstance(data, str):
        return False

    return data.startswith(rhs)


def lookup_istartswith(
    data: Union[str, list[str], Mapping[str, str]],
    rhs: Union[str, list[str], Mapping[str, str], Pattern[str]],
) -> bool:
    if not isinstance(rhs, str) or not isinstance(data, str):
        return False

    return data.lower().startswith(rhs.lower())


def lookup_endswith(
    data: Union[str, list[str], Mapping[str, str]],
    rhs: Union[str, list[str], Mapping[str, str], Pattern[str]],
) -> bool:
    if not isinstance(rhs, str) or not isinstance(data, str):
        return False

    return data.endswith(rhs)


def lookup_iendswith(
    data: Union[str, list[str], Mapping[str, str]],
    rhs: Union[str, list[str], Mapping[str, str], Pattern[str]],
) -> bool:
    if not isinstance(rhs, str) or not isinstance(data, str):
        return False
    return data.lower().endswith(rhs.lower())


def lookup_in(
    data: Union[str, list[str], Mapping[str, str]],
    rhs: Union[str, list[str], Mapping[str, str], Pattern[str]],
) -> bool:
    if isinstance(rhs, list):
        return data in rhs

    try:
        if isinstance(rhs, str) and isinstance(data, Mapping):
            return rhs in data
        if isinstance(rhs, str) and isinstance(data, (str, list)):
            return rhs in data
        if isinstance(rhs, str) and isinstance(data, Mapping):
            return rhs in data
        # TODO: Add a deep Mappingionary matcher
        # if isinstance(rhs, Mapping) and isinstance(data, Mapping):
        #     return rhs.items() not in data.items()
    except Exception:
        return False
    return False


def lookup_nin(
    data: Union[str, list[str], Mapping[str, str]],
    rhs: Union[str, list[str], Mapping[str, str], Pattern[str]],
) -> bool:
    if isinstance(rhs, list):
        return data not in rhs

    try:
        if isinstance(rhs, str) and isinstance(data, Mapping):
            return rhs not in data
        if isinstance(rhs, str) and isinstance(data, (str, list)):
            return rhs not in data
        if isinstance(rhs, str) and isinstance(data, Mapping):
            return rhs not in data
        # TODO: Add a deep Mappingionary matcher
        # if isinstance(rhs, Mapping) and isinstance(data, Mapping):
        #     return rhs.items() not in data.items()
    except Exception:
        return False
    return False


def lookup_regex(
    data: Union[str, list[str], Mapping[str, str]],
    rhs: Union[str, list[str], Mapping[str, str], Pattern[str]],
) -> bool:
    if isinstance(data, (str, bytes, re.Pattern)) and isinstance(rhs, (str, bytes)):
        return bool(re.search(rhs, data))
    return False


def lookup_iregex(
    data: Union[str, list[str], Mapping[str, str]],
    rhs: Union[str, list[str], Mapping[str, str], Pattern[str]],
) -> bool:
    if isinstance(data, (str, bytes, re.Pattern)) and isinstance(rhs, (str, bytes)):
        return bool(re.search(rhs, data, re.IGNORECASE))
    return False


LOOKUP_NAME_MAP: Mapping[str, LookupProtocol] = {
    "eq": lookup_exact,
    "exact": lookup_exact,
    "iexact": lookup_iexact,
    "contains": lookup_contains,
    "icontains": lookup_icontains,
    "startswith": lookup_startswith,
    "istartswith": lookup_istartswith,
    "endswith": lookup_endswith,
    "iendswith": lookup_iendswith,
    "in": lookup_in,
    "nin": lookup_nin,
    "regex": lookup_regex,
    "iregex": lookup_iregex,
}


class QueryList(list[T]):
    """Filter list of object/dictionaries. For small, local datasets.

    *Experimental, unstable*.

    >>> query = QueryList(
    ...     [
    ...         {
    ...             "place": "Largo",
    ...             "city": "Tampa",
    ...             "state": "Florida",
    ...             "foods": {"fruit": ["banana", "orange"], "breakfast": "cereal"},
    ...         },
    ...         {
    ...             "place": "Chicago suburbs",
    ...             "city": "Elmhurst",
    ...             "state": "Illinois",
    ...             "foods": {"fruit": ["apple", "cantelope"], "breakfast": "waffles"},
    ...         },
    ...     ]
    ... )
    >>> query.filter(place="Chicago suburbs")[0]['city']
    'Elmhurst'
    >>> query.filter(place__icontains="chicago")[0]['city']
    'Elmhurst'
    >>> query.filter(foods__breakfast="waffles")[0]['city']
    'Elmhurst'
    >>> query.filter(foods__fruit__in="cantelope")[0]['city']
    'Elmhurst'
    >>> query.filter(foods__fruit__in="orange")[0]['city']
    'Tampa'

    Examples of object lookups:

    >>> from typing import Any
    >>> from dataclasses import dataclass, field

    >>> @dataclass()
    ... class Restaurant:
    ...     place: str
    ...     city: str
    ...     state: str
    ...     foods: dict[str, Any]

    >>> restaurant = Restaurant(
    ...     place="Largo",
    ...     city="Tampa",
    ...     state="Florida",
    ...     foods={
    ...         "fruit": ["banana", "orange"], "breakfast": "cereal"
    ...     }
    ... )

    >>> restaurant
    Restaurant(place='Largo',
        city='Tampa',
        state='Florida',
        foods={'fruit': ['banana', 'orange'], 'breakfast': 'cereal'})

    >>> query = QueryList([restaurant])

    >>> query.filter(foods__fruit__in="banana")
    [Restaurant(place='Largo',
        city='Tampa',
        state='Florida',
        foods={'fruit': ['banana', 'orange'], 'breakfast': 'cereal'})]

    >>> query.filter(foods__fruit__in="banana")[0].city
    'Tampa'

    Example of deeper object lookups:

    >>> from typing import Optional
    >>> from dataclasses import dataclass, field

    >>> @dataclass()
    ... class Menu:
    ...     fruit: list[str] = field(default_factory=list)
    ...     breakfast: Optional[str] = None


    >>> @dataclass()
    ... class Restaurant:
    ...     place: str
    ...     city: str
    ...     state: str
    ...     menu: Menu = field(default_factory=Menu)


    >>> restaurant = Restaurant(
    ...     place="Largo",
    ...     city="Tampa",
    ...     state="Florida",
    ...     menu=Menu(
    ...         fruit=["banana", "orange"], breakfast="cereal"
    ...     )
    ... )

    >>> restaurant
    Restaurant(place='Largo',
        city='Tampa',
        state='Florida',
        menu=Menu(fruit=['banana', 'orange'], breakfast='cereal'))

    >>> query = QueryList([restaurant])

    >>> query.filter(menu__fruit__in="banana")
    [Restaurant(place='Largo',
        city='Tampa',
        state='Florida',
        menu=Menu(fruit=['banana', 'orange'], breakfast='cereal'))]

    >>> query.filter(menu__fruit__in="banana")[0].city
    'Tampa'
    """

    data: Sequence[T]
    pk_key: Optional[str]

    def items(self) -> list[T]:
        data: Sequence[T]

        if self.pk_key is None:
            raise Exception("items() require a pk_key exists")
        return [(getattr(item, self.pk_key), item) for item in self]

    def __eq__(
        self,
        other: object,
        # other: Union[
        #     "QueryList[T]",
        #     List[Mapping[str, str]],
        #     List[Mapping[str, int]],
        #     List[Mapping[str, Union[str, Mapping[str, Union[List[str], str]]]]],
        # ],
    ) -> bool:
        data = other

        if not isinstance(self, list) or not isinstance(data, list):
            return False

        if len(self) == len(data):
            for (a, b) in zip(self, data):
                if isinstance(a, Mapping):
                    a_keys = a.keys()
                    if a.keys == b.keys():
                        for key in a_keys:
                            if abs(a[key] - b[key]) > 1:
                                return False
                else:
                    if a != b:
                        return False

            return True
        return False

    def filter(
        self, matcher: Optional[Union[Callable[[T], bool], T]] = None, **kwargs: Any
    ) -> "QueryList[T]":
        def filter_lookup(obj: Any) -> bool:
            for path, v in kwargs.items():
                try:
                    lhs, op = path.rsplit("__", 1)

                    if op not in LOOKUP_NAME_MAP:
                        raise ValueError(f"{op} not in LOOKUP_NAME_MAP")
                except ValueError:
                    lhs = path
                    op = "exact"

                assert op in LOOKUP_NAME_MAP
                path = lhs
                data = keygetter(obj, path)

                if data is None or not LOOKUP_NAME_MAP[op](data, v):
                    return False

            return True

        if callable(matcher):
            _filter = matcher
        elif matcher is not None:

            def val_match(obj: Union[str, list[Any]]) -> bool:
                if isinstance(matcher, list):
                    return obj in matcher
                else:
                    return obj == matcher

            _filter = val_match
        else:
            _filter = filter_lookup

        return self.__class__(k for k in self if _filter(k))
